# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-06-20 13:09
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_fido', '0007_attestation_data_not_null'),
    ]

    operations = [
        migrations.AddField(
            model_name='authenticator',
            name='credential_id_data',
            field=models.TextField(null=True),
        ),
    ]
