# -*- coding: utf-8 -*-
# Generated by Django 1.11.28 on 2020-02-20 04:41
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('django_fido', '0012_authenticator_label'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authenticator',
            name='label',
            field=models.TextField(blank=True, default='', max_length=255),
            preserve_default=False,
        ),
        migrations.AlterUniqueTogether(
            name='authenticator',
            unique_together=set([('user', 'label')]),
        ),
    ]
