# Generated by Django 2.2.13 on 2020-07-22 11:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_fido', '0014_authenticator_metadata'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authenticatormetadata',
            name='url',
            field=models.URLField(unique=True),
        ),
    ]
